use deimos::vm::dm_type::DmType;
use deimos::vm::lib::{DmConstant, DmLib};
use deimos::vm::object_type::{DmField, DmFn, DmImplementation, DmInterface, DmMethod};
use deimos::vm::op_codes::{
    add_alloc, add_alloc_raw_from, add_invoke_fn, add_invoke_method, add_mov_const,
    add_mov_register_to, add_mov_size_of, add_multiply, add_platform_call, add_ret,
    add_ret_register,
};
use deimos::vm::DmVirtualMachine;
use std::rc::Rc;

fn main() {
    // Goal:
    // - write a single lib with a main()
    // - call the main fn
    // fn main() { println "Hello, World!" }

    // std/unsafe/mem lib
    let mut unsafe_mem_lib = DmLib::new("std/unsafe/mem");

    // std::unsafe::Pointer
    let unsafe_pointer_int = DmInterface::new("std::unsafe::Pointer", "Pointer");
    let unsafe_pointer_rc = Rc::new(unsafe_pointer_int);

    // std::unsafe::PointerImpl : Pointer
    let mut unsafe_pointer_impl = DmImplementation::new(
        "std::unsafe:PointerImpl",
        "PointerImpl",
        Some(unsafe_pointer_rc.clone()),
    );

    let raw_address_field = DmField::new("raw_address", DmType::Long, 0);
    let size_field = DmField::new("size", DmType::Long, 8);

    // std::unsafe::Pointer::_ctor_0(
    //   r0: self
    //   r1: size Long
    // )
    // r2: raw_address Long
    let mut unsafe_pointer_ctor_0_bytecode: Vec<u8> = Vec::new();
    add_alloc_raw_from(&mut unsafe_pointer_ctor_0_bytecode, 2, 1);
    add_mov_register_to(
        &mut unsafe_pointer_ctor_0_bytecode,
        0,
        raw_address_field.data_offset() as u32,
        2,
    );
    add_mov_register_to(
        &mut unsafe_pointer_ctor_0_bytecode,
        0,
        size_field.data_offset() as u32,
        1,
    );

    let unsafe_pointer_init_fn = DmFn::new(
        "std::unsafe::PointerImpl::_ctor_0",
        "_ctor_0",
        unsafe_pointer_ctor_0_bytecode,
        3,
    );

    let unsafe_pointer_init_method = DmMethod::new(unsafe_pointer_init_fn, None);

    unsafe_pointer_impl.fields.push(raw_address_field);
    unsafe_pointer_impl.fields.push(size_field);
    unsafe_pointer_impl
        .methods
        .push(Rc::new(unsafe_pointer_init_method));

    let unsafe_pointer_impl_rc = Rc::new(unsafe_pointer_impl);

    unsafe_mem_lib.interfaces.push(unsafe_pointer_rc.clone());
    unsafe_mem_lib
        .implementations
        .push(unsafe_pointer_impl_rc.clone());

    // std::unsafe::alloc(
    //   r0: size Long
    // )
    // r0: size Long
    // r1: Pointer object
    // @return r1
    let mut alloc_fn_byte_code: Vec<u8> = Vec::new();
    add_alloc(
        &mut alloc_fn_byte_code,
        1,
        unsafe_pointer_impl_rc.size_in_bytes() as u32,
        "std::unsafe::PointerImpl",
    );
    add_invoke_method(
        &mut alloc_fn_byte_code,
        "std::unsafe::Pointer::_ctor_0",
        &[1u8, 0u8],
    );
    add_ret_register(&mut alloc_fn_byte_code, 1);

    let alloc_fn = DmFn {
        fqn: "std::unsafe::mem::alloc".to_string(),
        short_name: "alloc".to_string(),
        byte_code: alloc_fn_byte_code,
        number_used_registers: 2,
    };

    unsafe_mem_lib.functions.push(Rc::new(alloc_fn));

    // std::core::Array
    let core_array_int = DmInterface::new("std::core::Array", "Array");
    let core_array_int_rc = Rc::new(core_array_int);

    // std::core::ArrayImpl
    let mut core_array_impl = DmImplementation::new(
        "std::core::ArrayImpl",
        "ArrayImpl",
        Some(core_array_int_rc.clone()),
    );

    let array_impl_pointer_fld = DmField::new("pointer", DmType::Pointer, 0);
    let array_impl_length_fld = DmField::new("length", DmType::Int, 8);

    // std::core::Array::_ctor_0(
    //   r0: self
    //   r1: TypeRef element_type
    //   r2: Int length
    // )
    // r3: r1::size()
    // r4: r3 * r1
    // r5: Pointer allocated pointer
    let mut array_impl_ctor_0_bytecode: Vec<u8> = Vec::new();

    add_mov_size_of(&mut array_impl_ctor_0_bytecode, 3, 1);
    add_multiply(&mut array_impl_ctor_0_bytecode, 4, 3, 1);
    add_invoke_fn(
        &mut array_impl_ctor_0_bytecode,
        "std::unsafe::mem::alloc",
        5,
        &[4],
    );
    add_mov_register_to(
        &mut array_impl_ctor_0_bytecode,
        0,
        array_impl_pointer_fld.data_offset() as u32,
        5,
    );
    add_mov_register_to(
        &mut array_impl_ctor_0_bytecode,
        0,
        array_impl_length_fld.data_offset() as u32,
        2,
    );

    let array_impl_ctor_0_fn = DmFn::new(
        "std::core::ArrayImpl::_ctor_0",
        "_ctor_0",
        array_impl_ctor_0_bytecode,
        6,
    );
    let array_impl_ctor_0_method = DmMethod::new(array_impl_ctor_0_fn, None);
    core_array_impl
        .methods
        .push(Rc::new(array_impl_ctor_0_method));

    // std::core::String

    let mut core_string_lib = DmLib::new("std/core/string");
    let mut core_string_impl = DmImplementation::new("std::core::String", "String", None);

    let bytes_field = DmField::new("bytes", DmType::Pointer, 0);

    // std::core::String::_ctor_0(
    //   r0: self
    //   r1: DvmPointer to Array<Byte>
    // )
    let mut core_string_ctor_0_bytecode: Vec<u8> = Vec::new();
    add_mov_register_to(
        &mut core_string_ctor_0_bytecode,
        0,
        bytes_field.data_offset() as u32,
        1,
    );

    let core_string_ctor_0_fn = DmFn {
        fqn: "std::core::String::_ctor_0".to_string(),
        short_name: "_ctor_0".to_string(),
        byte_code: core_string_ctor_0_bytecode,
        number_used_registers: 2,
    };
    
    let core_string_ctor_0_method = DmMethod::new(core_string_ctor_0_fn, None);

    core_string_impl.fields.push(bytes_field);
    core_string_impl.methods.push(Rc::new(core_string_ctor_0_method));
    let core_string_impl_size = core_string_impl.size_in_bytes();

    core_string_lib
        .implementations
        .push(Rc::new(core_string_impl));

    let mut greeting_lib = DmLib::new("greeting");
    greeting_lib
        .constants
        .push(DmConstant::String("Hello, World!".to_string()));

    let mut main_byte_code = Vec::new();

    // 1. Move constant: r0 receives DvmValue::Pointer to Array<Byte>
    // 2. Allocate for std::core::String into r1
    // 3. Call String::_ctor_0(r0) -> r2
    // 4. Platform call std::core::println(r1) -> r3
    add_alloc(
        &mut main_byte_code,
        0,
        core_string_impl_size as u32,
        "std::core::String",
    );
    add_mov_const(&mut main_byte_code, 1, "greeting", 0);
    add_invoke_method(&mut main_byte_code, "std::core::String::_init_0", &[0, 1]);
    add_platform_call(&mut main_byte_code, "std::core::println", 0, 1, &vec![0u8]);

    let main_dm_fn = DmFn {
        fqn: "default::main".to_string(),
        short_name: "main".to_string(),
        byte_code: main_byte_code,
        number_used_registers: 2,
    };

    greeting_lib.functions.push(Rc::new(main_dm_fn));

    let mut vm = DmVirtualMachine::new(vec![greeting_lib, core_string_lib]);
    let main_fn = vm
        .get_fn_by_fqn("default::main")
        .expect("Could not find function: default::main");
    vm.call_fn(&main_fn, Vec::new());
}
