pub mod dm_type;
pub mod dvm_value;
pub mod lib;
pub mod mem;
pub mod object_type;
pub mod op_codes;
pub mod platform;
pub mod util;

use crate::vm::dvm_value::DvmValue;
use crate::vm::lib::{DmConstant, DmLib};
use crate::vm::mem::{get_field_value, DmAllocObject};
use crate::vm::object_type::DmFn;
use crate::vm::platform::init_platform_functions;
use op_codes::*;
use std::alloc::{alloc, dealloc, Layout};
use std::collections::HashMap;
use std::rc::Rc;
use util::{get_32_le, get_64_le};

pub type PlatformFunction = fn(args: Vec<DvmValue>, &mut DmVirtualMachine) -> DvmValue;

enum CallFrame {
    PlatformCall(CallFrameInfo),
    DeimosCall(CallFrameInfo),
}

struct CallFrameInfo {
    fn_fqn: String,
}

#[deprecated]
pub struct DmVirtualMachine {
    libs: Vec<DmLib>,
    functions: HashMap<String, Rc<DmFn>>,
    platform_functions: HashMap<String, PlatformFunction>,
    ip: usize,
    call_stack: Vec<CallFrame>,
    registers: Vec<DvmValue>,
    register_state_stack: Vec<Vec<DvmValue>>,
}

pub struct DvmContext {
    libs: Vec<DmLib>,
    functions: HashMap<String, Rc<DmFn>>,
    platform_functions: HashMap<String, Rc<PlatformFunction>>,
}

impl DvmContext {
    pub fn new() -> Self {
        DvmContext {
            libs: Vec::new(),
            functions: HashMap::new(),
            platform_functions: HashMap::new(),
        }
    }

    pub fn fn_by_fqn(&self, fqn: &str) -> Option<Rc<DmFn>> {
        self.functions.get(fqn).cloned()
    }

    pub fn platform_fn_by_fqn(&self, fqn: &str) -> Option<Rc<PlatformFunction>> {
        self.platform_functions.get(fqn).cloned()
    }
}

pub struct DvmState {
    call_stack: Vec<CallFrame>,
    registers: Vec<DvmValue>,
    register_state_stack: Vec<Vec<DvmValue>>,
}

impl DvmState {
    pub fn new() -> Self {
        DvmState {
            call_stack: Vec::new(),
            registers: Vec::new(),
            register_state_stack: Vec::new(),
        }
    }
}

fn load_functions(destination: &mut HashMap<String, Rc<DmFn>>, libs: &Vec<DmLib>) {
    for lib in libs {
        for lib_fn in &lib.functions {
            destination.insert(lib_fn.fqn.clone(), lib_fn.clone());
        }
        for interface_function in lib
            .interfaces
            .iter()
            .flat_map(|interface| interface.get_functions())
        {
            destination.insert(interface_function.fqn.clone(), interface_function.clone());
        }
        for implementation_function in lib
            .implementations
            .iter()
            .flat_map(|implementation| &implementation.functions)
        {
            destination.insert(
                implementation_function.fqn.clone(),
                implementation_function.clone(),
            );
        }
    }
}

pub fn call_fn(
    state: &mut DvmState,
    context: &DvmContext,
    dm_fn: &DmFn,
    args: Vec<DvmValue>,
) -> Option<DvmValue> {
    // save current state
    state.call_stack.push(CallFrame::DeimosCall(CallFrameInfo {
        fn_fqn: dm_fn.fqn.clone(),
    }));
    state.register_state_stack.push(state.registers.clone());

    // zero registers and make sure there are enough for dm_function
    state.registers.clear();
    if state.registers.len() < dm_fn.number_used_registers() {
        state
            .registers
            .resize(dm_fn.number_used_registers(), DvmValue::Uninit);
    }

    // push args
    for arg in args {
        state.registers.push(arg);
    }

    // run the byte code
    run_byte_code(state, context, &dm_fn.byte_code);

    // get return value
    let return_value = state
        .registers
        .get(dm_fn.return_register() as usize)
        .map(|m| m.clone());

    // restore state
    state.registers = state.register_state_stack.pop().unwrap();
    state.call_stack.pop();

    // return result
    return_value
}

macro_rules! next_8 {
    ( $byte_code: expr, $T: ident ) => {
        $byte_code.next().unwrap() as $T
    };
}

macro_rules! next_32_le {
    ( $byte_code: expr, $T: ident ) => {
        next_8!($byte_code, $T)
            + (next_8!($byte_code, $T) << 8)
            + (next_8!($byte_code, $T) << 16)
            + (next_8!($byte_code, $T) << 24)
    };
}

macro_rules! next_64_le {
    ( $byte_code: expr, $T: ident ) => {
        next_8!($byte_code, $T)
            + (next_8!($byte_code, $T) << 8)
            + (next_8!($byte_code, $T) << 16)
            + (next_8!($byte_code, $T) << 24)
            + (next_8!($byte_code, $T) << 32)
            + (next_8!($byte_code, $T) << 40)
            + (next_8!($byte_code, $T) << 48)
            + (next_8!($byte_code, $T) << 56)
    };
}

macro_rules! read_string {
    ( $iter: expr ) => {{
        let str_length = next_32_le!($iter, usize);
        let mut str_raw: Vec<u8> = Vec::new();
        let mut i = 0;
        while i < str_length {
            str_raw.push($iter.next().unwrap());
            i += 1;
        }
        String::from_utf8(str_raw).unwrap()
    }};
}

macro_rules! write_string {
    ( $raw_ptr: expr, $s: expr ) => {
        for (i, b) in $s.bytes().enumerate() {
            unsafe {
                $raw_ptr.add(i).write(b);
            }
        }
    };
}

macro_rules! write_bytes {
    ( $ptr: expr, $offset: expr, $src: expr ) => {
        for (i, &b) in $src.iter().enumerate() {
            unsafe {
                $ptr.add($offset + i).write(b);
            }
        }
    };
}

pub fn run_byte_code(state: &mut DvmState, context: &DvmContext, byte_code: &[u8]) {
    let mut iter = byte_code.iter().cloned();
    while let Some(op_code) = iter.next() {
        match op_code {
            MOV_INT => {
                let target_register = next_8!(iter, usize);
                let operand = next_32_le!(iter, u32) as i32;
                state
                    .registers
                    .insert(target_register, DvmValue::Int(operand));
            }
            MOV_LONG => {
                let target_register = next_8!(iter, usize);
                let operand = next_64_le!(iter, u64) as i64;
                state
                    .registers
                    .insert(target_register, DvmValue::Long(operand));
            }
            MOV_DOUBLE => {
                let target_register = next_8!(iter, usize);
                let operand = next_64_le!(iter, u64) as f64;
                state
                    .registers
                    .insert(target_register, DvmValue::Double(operand));
            }
            MOV_REGISTER => {
                let target_register = next_8!(iter, usize);
                let source_register = next_8!(iter, usize);
                let source_value = state.registers.get(source_register).unwrap();
                state
                    .registers
                    .insert(target_register, source_value.clone())
            }
            ALLOC => {
                let target_register = next_8!(iter, usize);
                let alloc_size = next_32_le!(iter, usize);
                let impl_name = read_string!(iter);

                let implementation = context
                    .libs
                    .iter()
                    .find_map(|lib| {
                        lib.implementations
                            .iter()
                            .find(|implementation| implementation.fqn == impl_name)
                    })
                    .expect(&format!("Implementation not found: {}", impl_name));

                let dm_alloc_object = DmAllocObject::new(alloc_size, implementation.clone());
                let dvm_pointer = DvmValue::Pointer(Rc::new(dm_alloc_object));
                state.registers.insert(target_register, dvm_pointer);
            }
            DEALLOC => {
                let target_register = next_8!(iter, usize);
                let dm_alloc_object = state.registers
                    .remove(target_register)
                    .expect_pointer();
                drop(dm_alloc_object); // explicit
            }
            MOV_INT_TO => {
                let target_register = next_8!(iter, usize);
                let offset = next_32_le!(iter, usize);
                let operand = next_32_le!(iter, u32) as i32;

                let dm_alloc_object = state
                    .registers
                    .get(target_register)
                    .unwrap()
                    .expect_pointer();
                write_bytes!(dm_alloc_object.data, offset, operand.to_ne_bytes());
            }
            MOV_LONG_TO => {
                todo!()
            }
            MOV_DOUBLE_TO => {
                todo!()
            }
            MOV_REGISTER_TO => {
                // Moves the value from the source register to the object pointed to by the target
                // register, at the given offset.
                let target_register = next_8!(iter, usize);
                let offset = next_32_le!(iter, usize);
                let source_register = next_8!(iter, usize);

                let target_alloc_object = state
                    .registers
                    .get(target_register)
                    .unwrap()
                    .expect_pointer();

                let source_value = state.registers.get(source_register).unwrap();
                match source_value {
                    DvmValue::Int(i) => {
                        write_bytes!(target_alloc_object.data, offset, i.to_ne_bytes());
                    }
                    DvmValue::Byte(b) => unsafe {
                        target_alloc_object.data.add(offset).write(*b);
                    },
                    DvmValue::Long(l) => {
                        write_bytes!(target_alloc_object.data, offset, l.to_ne_bytes());
                    }
                    DvmValue::Double(d) => {
                        write_bytes!(target_alloc_object.data, offset, d.to_ne_bytes());
                    }
                    DvmValue::Boolean(b) => unsafe {
                        target_alloc_object.data.add(offset).write(*b as u8);
                    },
                    DvmValue::Pointer(source_alloc_object) => {
                        let source_ptr = Rc::into_raw(source_alloc_object.clone());
                        write_bytes!(
                            target_alloc_object.data,
                            offset,
                            (source_ptr as usize).to_ne_bytes()
                        );
                    }
                    DvmValue::Uninit => {
                        panic!("Cannot move DvmValue::Uninit to object.")
                    }
                }
            }
            MOV_CONST => {
                // TODO: reduce the complexity of this.
                // Perhaps the target register should just contain a ptr where the data is written.

                // Decode
                let target_register = next_8!(iter, usize); // must contain an allocated StringImpl
                let lib_name = read_string!(iter);
                let const_id = next_8!(iter, usize);
                let pointer_impl_register = next_8!(iter, usize);
                let byte_array_impl_register = next_8!(iter, usize);

                // Get constant
                let Some(lib) = context.libs.iter().find(|lib| lib.name == lib_name) else {
                    panic!("Could not find lib with name: {}", lib_name);
                };
                let constant = lib.constants.get(const_id).unwrap();
                let DmConstant::String(s) = constant;

                // Get allocated PointerImpl
                let pointer_impl = state
                    .registers
                    .get(pointer_impl_register)
                    .unwrap()
                    .expect_pointer();

                // Init PointerImpl object
                let pointer_impl_ctor_0_fn = context
                    .functions
                    .get("std::unsafe::PointerImpl::_ctor_0")
                    .expect("Could not find std::unsafe::PointerImpl::_ctor_0");
                call_fn(
                    state,
                    context,
                    pointer_impl_ctor_0_fn,
                    vec![
                        DvmValue::Pointer(pointer_impl.clone()),
                        DvmValue::Long(s.len() as i64),
                    ],
                );

                // Get std::unsafe::PointerImpl.raw_address field as *mut u8
                let raw_ptr = unsafe {
                    get_field_value(
                        pointer_impl
                            .implementation
                            .get_field("raw_address", &pointer_impl)
                            .expect("Could not get PointerImpl.raw_address field."),
                        &pointer_impl,
                    )
                }
                .expect_long() as usize as *mut u8;

                // Write the constant bytes to the raw_ptr
                write_string!(raw_ptr, s);

                // Get allocated ArrayImpl<Byte> from byte_array_impl_register
                let byte_array_impl = state
                    .registers
                    .get(byte_array_impl_register)
                    .unwrap()
                    .expect_pointer();

                // Init ArrayImpl<Byte>
                let array_ctor_0_fn = context
                    .functions
                    .get("std::core::ArrayImpl::_ctor_1")
                    .expect("Could not find std::core::ArrayImpl::_ctor_1");
                call_fn(
                    state,
                    context,
                    array_ctor_0_fn,
                    vec![
                        DvmValue::Pointer(byte_array_impl.clone()), // self
                        DvmValue::Pointer(pointer_impl.clone()),    // PointerImpl
                        DvmValue::Int(s.len() as i32),              // length
                    ],
                );

                // Get allocated StringImpl from string_impl_register
                let string_impl = state
                    .registers
                    .get(target_register)
                    .unwrap()
                    .expect_pointer();

                // Init StringImpl
                let string_impl_ctor_0_fn = context
                    .functions
                    .get("std::core::StringImpl::_ctor_0")
                    .expect("Could not find std::core::StringImpl::_ctor_0");
                call_fn(
                    state,
                    context,
                    string_impl_ctor_0_fn,
                    vec![
                        DvmValue::Pointer(string_impl.clone()),
                        DvmValue::Pointer(byte_array_impl.clone()),
                    ],
                );
            }
            ALLOC_RAW => {
                // Allocates a raw number of bytes, with the number of bytes determined by the
                // operand
                todo!()
            }
            ALLOC_RAW_FROM => {
                // Allocates a raw number of bytes, with the number of bytes determined by the value
                // of the source register
                todo!()
            }
            PLATFORM_CALL => {
                todo!()
            }
            INVOKE_FN => {
                todo!()
            }
            INVOKE_VIRTUAL => {
                todo!()
            }
            INVOKE_DYNAMIC => {
                todo!()
            }
            RETURN_REGISTER => {
                todo!()
            }
            MOV_SIZE_OF => {
                todo!()
            }
            MULTIPLY => {
                todo!()
            }
            op_code => {
                panic!("Invalid op_code: {}", op_code);
            }
        }
    }
}

impl DmVirtualMachine {
    pub fn new(libs: Vec<DmLib>) -> DmVirtualMachine {
        let mut vm = DmVirtualMachine {
            libs: libs,
            functions: HashMap::new(),
            platform_functions: init_platform_functions(),
            ip: 0,
            registers: Vec::new(),
            call_stack: Vec::new(),
            register_state_stack: Vec::new(),
        };
        load_functions(&mut vm.functions, &vm.libs);
        vm
    }

    #[deprecated]
    pub fn get_fn_by_fqn(&self, fqn: &str) -> Option<Rc<DmFn>> {
        self.functions.get(fqn).cloned()
    }

    #[deprecated]
    pub fn call_fn(&mut self, dm_function: &DmFn, args: Vec<DvmValue>) -> Option<DvmValue> {
        // save current state
        self.call_stack.push(CallFrame::DeimosCall(DeimosCallFrame {
            return_address: self.ip,
        }));
        self.register_state_stack.push(self.registers.clone());

        // zero registers and make sure there are enough for dm_function
        self.registers.clear();
        if self.registers.len() < args.len() {
            self.registers.resize(args.len(), DvmValue::Uninit);
        }

        // push args
        for i in 0..args.len() {
            self.registers.insert(i, args[i].clone());
        }

        // run the byte code
        self.run_raw(&dm_function.byte_code);

        // restore state
        self.registers = self.register_state_stack.pop().unwrap();
        if let CallFrame::DeimosCall(deimos_call_frame) = self.call_stack.pop().unwrap() {
            self.ip = deimos_call_frame.return_address;
        }

        // return result
        self.registers.get(0).map(|x| x.clone())
    }

    pub fn run_raw(&mut self, code: &Vec<u8>) {
        let mut i = 0;
        while i < code.len() {
            let op_code = code[i];
            println!("op_code: {:#04x}", op_code);
            match code[i] {
                MOV_INT => {
                    let target_register = code[i + 1] as usize;
                    let operand = next_32_le!(code, i, 2, u32);
                    self.registers
                        .insert(target_register, DvmValue::Int(operand as i32));
                    i += 6;
                }
                MOV_LONG => {
                    let target_register = code[i + 1] as usize;
                    let operand = get_64_le!(code, i, 2, u64);
                    self.registers
                        .insert(target_register, DvmValue::Long(operand as i64));
                    i += 10;
                }
                MOV_DOUBLE => {
                    unimplemented!();
                }
                MOV_REGISTER => {
                    let target_register = code[i + 1] as usize;
                    let source_register = code[i + 2] as usize;
                    let source_value = self.registers.get(source_register).unwrap();
                    self.registers.insert(target_register, source_value.clone());
                    i += 3;
                }
                ALLOC => {
                    i += 1;
                    let target_register = code[i] as usize;
                    i += 1;
                    let alloc_size = next_32_le!(code, i, 0, usize);
                    i += 4;
                    let raw_implementation_name_length = next_32_le!(code, i, 0, usize);
                    i += 4;
                    let raw_implementation_name_bytes =
                        code[i..(i + raw_implementation_name_length)].to_vec();
                    i += raw_implementation_name_length;
                    let implementation_name =
                        String::from_utf8(raw_implementation_name_bytes).unwrap();

                    let implementation = self
                        .libs
                        .iter()
                        .find_map(|lib| {
                            lib.implementations
                                .iter()
                                .find(|implementation| implementation.fqn == implementation_name)
                        })
                        .expect(&format!(
                            "Implementation not found: {}",
                            implementation_name
                        ));

                    let layout = Layout::from_size_align(alloc_size, 1).unwrap();
                    let raw_data_ptr;
                    unsafe {
                        raw_data_ptr = alloc(layout);
                    }
                    let dm_alloc_object = DmAllocObject {
                        data: raw_data_ptr,
                        layout,
                        size: alloc_size,
                        implementation: implementation.clone(),
                    };

                    let dvm_value = DvmValue::Pointer(Rc::new(dm_alloc_object));
                    self.registers.insert(target_register, dvm_value);
                }
                DEALLOC => {
                    todo!()
                }
                MOV_INT_TO => {
                    todo!()
                }
                MOV_REGISTER_TO => {
                    i += 1;
                    let target_register = code[i] as usize;
                    i += 1;
                    let offset = next_32_le!(code, i, 0, isize);
                    i += 4;
                    let source_register = code[i] as usize;
                    i += 1;

                    let target_value = self.registers.get(target_register).unwrap();
                    let target_pointer = match target_value {
                        DvmValue::Pointer(alloc_object) => alloc_object.data,
                        _ => panic!("Expected a DvmValue::Pointer, but got: {:?}", target_value),
                    };

                    let source_value = self.registers.get(source_register).unwrap();
                    match source_value {
                        DvmValue::Pointer(source_alloc_object) => unsafe {
                            let source_pointer = Rc::into_raw(source_alloc_object.clone());
                            for (j, b) in (source_pointer as usize).to_le_bytes().iter().enumerate()
                            {
                                target_pointer.offset(offset + j as isize).write(*b);
                            }
                        },
                        _ => panic!("Currently unsupported source value: {:?}", source_value),
                    }
                }
                MOV_CONST => {
                    i += 1;
                    let target_register = code[i] as usize;
                    i += 1;
                    let lib_name_length = next_32_le!(code, i, 0, usize);
                    i += 4;
                    let lib_name_raw = code[i..(i + lib_name_length)].to_vec();
                    i += lib_name_length;
                    let const_id = next_32_le!(code, i, 0, usize);
                    i += 4;

                    let lib_name = String::from_utf8(lib_name_raw).unwrap();

                    if let Some(lib) = self.libs.iter().find(|&lib| lib.name == lib_name) {
                        let constant = &lib.constants[const_id];
                        match constant {
                            DmConstant::String(s) => {
                                let alloc_fn = self
                                    .get_fn_by_fqn("std::unsafe::mem::alloc")
                                    .expect("Could not find std::unsafe::mem::alloc");
                                let alloc_size_arg = DvmValue::Long(s.len() as i64);
                                let alloc_return_value =
                                    self.call_fn(&alloc_fn, vec![alloc_size_arg]).unwrap();
                                let DvmValue::Pointer(pointer_object) = alloc_return_value else {
                                    panic!("Expected std::unsafe::mem::alloc to return DvmValue::Pointer, but got: {:?}", alloc_return_value);
                                };
                                let raw_address_field = pointer_object
                                    .implementation
                                    .get_field("raw_address", &pointer_object)
                                    .expect("Could not get PointerImpl.raw_address field.");
                                unsafe {
                                    let raw_address_value =
                                        get_field_value(&raw_address_field, &pointer_object);
                                    let DvmValue::Long(raw_address) = raw_address_value else {
                                        panic!("Expected PointerImpl.raw_address to be a DvmValue::Long, but got: {:?}", raw_address_value);
                                    };
                                    let ptr = raw_address as usize as *mut u8;
                                    for (j, b) in s.bytes().enumerate() {
                                        ptr.offset(j as isize).write(b);
                                    }
                                    self.registers
                                        .insert(target_register, DvmValue::Pointer(pointer_object));
                                }
                            }
                            _ => {
                                panic!("Invalid constant type");
                            }
                        }
                    } else {
                        panic!("Could not find lib: {}", lib_name);
                    }
                }
                PLATFORM_CALL => {
                    i += 1;

                    let symbol_name_length = next_32_le!(code, i, 0, usize);
                    i += 4;

                    let symbol_name_raw = code[i..(i + symbol_name_length)].to_vec();
                    i += symbol_name_length;

                    let symbol_name = String::from_utf8(symbol_name_raw).unwrap();

                    let return_register = code[i + 1] as usize;
                    i += 1;

                    let arg_registers_length = code[i] as usize;
                    i += 1;

                    let arg_registers = code[i..(i + arg_registers_length)].to_vec();
                    i += arg_registers_length;

                    let mut args = Vec::new();
                    for arg_register in arg_registers {
                        let register_value = self.registers.get(arg_register as usize).unwrap();
                        args.push(register_value.clone());
                    }

                    let platform_function_result = self.platform_functions.get(&symbol_name);
                    if platform_function_result.is_none() {
                        panic!("Unknown platform function {}", symbol_name);
                    }
                    let platform_function = platform_function_result.unwrap();

                    let call_result = platform_function(args, self);
                    self.registers.insert(return_register, call_result);
                }
                INVOKE_FN => {
                    unimplemented!()
                }
                INVOKE_VIRTUAL => {
                    unimplemented!();
                }
                INVOKE_METHOD => {
                    i += 1;
                    let name_length = next_32_le!(code, i, 0, usize);
                    i += 4;
                    let name_raw = code[i..(i + name_length)].to_vec();
                    i += name_length;
                    let arg_registers_length = code[i] as usize;
                    i += 1;
                    let arg_registers = code[i..(i + arg_registers_length)].to_vec();
                    i += arg_registers_length;

                    let method_name = String::from_utf8(name_raw).unwrap();

                    let method = self
                        .get_fn_by_fqn(&method_name)
                        .expect(&format!("Could not find method: {}", method_name));

                    let mut args: Vec<DvmValue> = Vec::new();
                    for arg_register in arg_registers {
                        let register_value = self.registers.get(arg_register as usize).unwrap();
                        args.push(register_value.clone());
                    }

                    self.call_fn(method.as_ref(), args);
                }
                INVOKE_DYNAMIC => {
                    unimplemented!()
                }
                RETURN => {
                    i += 1;
                }
                RETURN_REGISTER => {
                    unimplemented!()
                }
                op_code => panic!("Invalid or unimplemented op code: {:#04x}", op_code),
            }
        }
    }
}

#[cfg(test)]
mod dvm_run_tests {
    use super::*;

    macro_rules! assert_register {
        ( $expected: expr, $register: expr ) => {
            assert_eq!($expected, $register.unwrap().clone());
        };
    }

    #[test]
    fn mov_1_as_int() {
        let mut code = Vec::new();
        add_mov_int(&mut code, 0, 1);
        let mut vm = DmVirtualMachine::new(Vec::new());
        vm.run_raw(&code);
        assert_register!(DvmValue::Int(1), vm.registers.get(0));
    }

    #[test]
    fn move_65535_as_int() {
        let mut code = Vec::new();
        add_mov_int(&mut code, 0, 0xffff);
        let mut vm = DmVirtualMachine::new(Vec::new());
        vm.run_raw(&code);
        assert_register!(DvmValue::Int(0xffff), vm.registers.get(0));
    }

    #[test]
    fn move_int_max_as_int() {
        let mut code = Vec::new();
        add_mov_int(&mut code, 0, 0x0fff_ffff);
        let mut vm = DmVirtualMachine::new(Vec::new());
        vm.run_raw(&code);
        assert_register!(DvmValue::Int(0x0fff_ffff), vm.registers.get(0));
    }

    #[test]
    fn move_register() {
        let mut code = Vec::new();
        add_mov_int(&mut code, 1, 1);
        add_mov_register(&mut code, 0, 1);
        let mut vm = DmVirtualMachine::new(Vec::new());
        vm.registers.resize(2, DvmValue::Uninit);
        vm.run_raw(&code);
        assert_register!(DvmValue::Int(1), vm.registers.get(0));
    }

    #[test]
    #[ignore]
    fn mov_int_to_register_as_address() {
        let mut code = Vec::new();
        add_alloc(&mut code, 0, 4);
        add_mov_int_to(&mut code, 0, 0, 0xff);
        let mut vm = DmVirtualMachine::new(Vec::new());
        vm.run_raw(&code);
        let actual = vm.registers.get(0).unwrap().clone();
        match actual {
            DvmValue::Pointer(value_box) => {
                assert_eq!(DvmValue::Int(0xff), *value_box);
            }
            _ => panic!("Target register is not a pointer"),
        }
    }

    #[test]
    #[ignore]
    fn alloc_and_dealloc_expect_register_cleared() {
        let mut code = Vec::new();
        add_alloc(&mut code, 0, 4);
        add_dealloc(&mut code, 0);
        let mut vm = DmVirtualMachine::new(Vec::new());
        vm.run_raw(&code);
        assert_register!(DvmValue::Uninit, vm.registers.get(0));
    }
}
