use crate::vm::dvm_value::DvmValue;
use crate::vm::mem::{get_field_value, DmAllocObject};
use crate::vm::DmVirtualMachine;
use std::rc::Rc;

pub fn dm_print(args: Vec<DvmValue>, vm: &mut DmVirtualMachine) -> DvmValue {
    if args.len() != 1 {
        return DvmValue::Uninit; // TODO: make exception
    }
    unsafe {
        print!("{}", get_string(&args[0], vm));
    }
    DvmValue::Uninit
}

pub fn dm_println(args: Vec<DvmValue>, vm: &mut DmVirtualMachine) -> DvmValue {
    if args.len() != 1 {
        return DvmValue::Uninit;
    }
    unsafe {
        println!("{}", get_string(&args[0], vm));
    }
    DvmValue::Uninit
}

unsafe fn get_string(dvm_value: &DvmValue, vm: &mut DmVirtualMachine) -> String {
    match dvm_value {
        DvmValue::Byte(b) => b.to_string(),
        DvmValue::Int(i) => i.to_string(),
        DvmValue::Long(l) => l.to_string(),
        DvmValue::Double(d) => d.to_string(),
        DvmValue::Boolean(b) => b.to_string(),
        DvmValue::Pointer(alloc_object_rc) => convert_to_string(alloc_object_rc.clone(), vm),
        DvmValue::Uninit => String::from("Unit"),
    }
}

unsafe fn convert_to_string(
    alloc_object_rc: Rc<DmAllocObject>,
    vm: &mut DmVirtualMachine,
) -> String {
    if alloc_object_rc.implementation.fqn == "std::core::String" {
        let bytes_field = alloc_object_rc
            .implementation
            .get_field("bytes", &alloc_object_rc)
            .expect("Could not get String.bytes field.");
        let bytes_value = get_field_value(&bytes_field, &alloc_object_rc);
        match bytes_value {
            DvmValue::Pointer(bytes_object) => {
                if bytes_object.implementation.fqn == "std::core::ArrayImpl" {
                    let pointer_field = bytes_object
                        .implementation
                        .get_field("pointer", &bytes_object)
                        .expect("Could not get ArrayImpl.pointer field.");
                    let pointer_value = get_field_value(&pointer_field, &alloc_object_rc);
                    match pointer_value {
                        DvmValue::Pointer(pointer_object) => {
                            if pointer_object.implementation.fqn == "std::unsafe::PointerImpl" {
                                let raw_address_field = pointer_object
                                    .implementation
                                    .get_field("raw_address", &pointer_object)
                                    .expect("Could not get PointerImpl.raw_address field.");
                                let raw_address_value =
                                    get_field_value(raw_address_field, &alloc_object_rc);

                                let size_field = pointer_object
                                    .implementation
                                    .get_field("size", &alloc_object_rc)
                                    .expect("Could not get PointerImpl.size field.");
                                let size_value = get_field_value(&size_field, &alloc_object_rc);
                                let DvmValue::Long(size) = size_value else {
                                    panic!("Expected PointerImpl.size to be a Long");
                                };

                                match raw_address_value {
                                    DvmValue::Long(raw_address) => {
                                        let raw_bytes_pointer = raw_address as usize as *mut u8;
                                        let mut v: Vec<u8> = Vec::new();
                                        for i in 0..(size as isize) {
                                            v.push(raw_bytes_pointer.offset(i).read());
                                        }
                                        String::from_utf8(v).unwrap()
                                    }
                                    _ => panic!("Expected raw_address to be a Long."),
                                }
                            } else {
                                panic!("Expected ArrayImpl.pointer to be a Pointer to PointerImpl, instead found: {:?}", pointer_object);
                            }
                        }
                        _ => panic!(
                            "Expected ArrayImpl.pointer to be a Pointer, instead found: {:?}",
                            pointer_value
                        ),
                    }
                } else {
                    panic!("Expected String.bytes to be a Pointer to an ArrayImpl, instead found: {:?}", bytes_object.implementation.fqn);
                }
            }
            _ => panic!(
                "Expected String.bytes to be a Pointer, instead found: {:?}",
                bytes_value
            ),
        }
    } else {
        todo!("what happens if we don't have a String?")
    }
}
